const { expect } = require('chai')
const sinon = require('sinon')
const DependencyInjector = require('../src/DependencyInjector')

describe('the dependency injector', function() {
    it('should be a functional class', function() {
        expect(DependencyInjector).to.be.a('function')
    })

    it('should be constructable', function(){
        expect(new DependencyInjector()).to.be.an.instanceof(DependencyInjector)
    })

    it('should set the container on construction', function() {
        const fake_container = {}
        const di = new DependencyInjector(fake_container)
        expect(di.container).to.be.equal(fake_container)
    })

    it('should inject itself into the container', function() {
        const fake_container = {}
        const di = new DependencyInjector(fake_container)
        expect(fake_container.di).to.not.be.undefined
        expect(fake_container.di).to.be.equal(di)
    })

    describe('make function', function() {
        let container
        let di
        let fake_class
        beforeEach(function() {
            container = {}
            di = new DependencyInjector(container)
            di.__is_injectable = sinon.fake.returns(true)
            fake_class = { __inject: sinon.spy(), name: 'fake_class' }
        })

        it('should check if the class is injectable', function() {
            di.make(fake_class)
            expect(di.__is_injectable.callCount).to.be.equal(1)
        })

        it('should fail if the class is not injectable', function() {
            di.__is_injectable = sinon.fake.returns(false)
            expect(di.make.bind(di, fake_class)).to.throw('Cannot inject non-injectable class: '+fake_class.name)
        })

        it('should call the inject method on class', function() {
            di.make(fake_class)
            expect(fake_class.__inject.calledOnce).to.be.true
        })

        it('should pass the container to the inject method', function() {
            di.make(fake_class)
            expect(fake_class.__inject.getCall(0).args.length).to.be.equal(1)
            expect(fake_class.__inject.getCall(0).args[0]).to.be.equal(container)
        })

        it('should return the class', function() {
            expect(di.make(fake_class)).to.be.equal(fake_class)
        })
    })

    describe('service function', function() {
        let container
        let di
        beforeEach(function() {
            container = { service: sinon.spy() }
            di = new DependencyInjector(container)
        })

        it('should be a function', function() {
            expect(di.service).to.be.a('function')
        })

        it('should expect one argument', function() {
            expect(di.service.length).to.be.equal(1)
        })

        it('should call the service method on the container', function() {
            const arg = {}
            di.service(arg)
            expect(container.service.calledOnce).to.be.true
            expect(container.service.getCall(0).args.length).to.be.equal(1)
            expect(container.service.getCall(0).args[0]).to.be.equal(arg)
        })
    })

    describe('has function', function() {
        let container
        let di
        beforeEach(function() {
            container = { has: sinon.spy() }
            di = new DependencyInjector(container)
        })

        it('should be a function', function() {
            expect(di.has).to.be.a('function')
        })

        it('should expect 1 argument', function() {
            expect(di.has.length).to.be.equal(1)
        })

        it('should call the has method on the container', function() {
            const name = Math.random().toString(36).substring(7)
            di.has(name)
            expect(container.has.calledOnce).to.be.true
            expect(container.has.getCall(0).args.length).to.be.equal(1)
            expect(container.has.getCall(0).args[0]).to.be.equal(name)
        })

        it('should call the has method on the container for each element in the argument array, one is passed', function() {
            const n_items = parseInt(String(Math.random() * 100))
            const arr = Array(n_items).fill('')
            container.has = sinon.fake.returns(true)
            expect(di.has(arr)).to.be.true
            expect(container.has.getCalls().length).to.be.equal(n_items)
        })
    })

    describe('inject_globally function', function() {
        it('should be a function', function() {
            const di = new DependencyInjector()
            expect(di.inject_globally).to.be.a('function')
        })

        it('should accept no arguments', function() {
            const di = new DependencyInjector()
            expect(di.inject_globally.length).to.be.equal(0)
        })

        it('should replace the global require method', function() {
            const mod_proto = require('module').prototype
            const original_require = mod_proto.require
            const di = new DependencyInjector()
            di.inject_globally()
            expect(mod_proto.require).to.not.be.equal(original_require)
            mod_proto.require = original_require
        })
    })

    describe('__is_injectable function', function() {
        let di
        let SomeClass
        beforeEach(function() {
            di = new DependencyInjector()
            class TestClass {
                static __inject(a) {}
            }
            SomeClass = TestClass
        })

        it('should be a function', function() {
            expect(di.__is_injectable).to.be.a('function')
        })

        it('should accept one argument', function() {
            expect(di.__is_injectable.length).to.be.equal(1)
        })

        it('should fail if the class is not functional', function() {
            expect(di.__is_injectable(SomeClass)).to.be.true
            SomeClass = {}
            expect(di.__is_injectable(SomeClass)).to.be.false
        })

        it('should fail if the class does not have a static __inject function', function() {
            expect(di.__is_injectable(SomeClass)).to.be.true
            SomeClass.__inject = true
            expect(di.__is_injectable(SomeClass)).to.be.false
        })

        it('should fail if __inject does not take at least one argument', function() {
            expect(di.__is_injectable(SomeClass)).to.be.true
            SomeClass.__inject = function() {}
            expect(di.__is_injectable(SomeClass)).to.be.false
        })
    })
})
